/*
 * Copyright (c) 2005-2014, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Skanderborgvej 40 4-2
 * DK-8000 Aarhus C
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#ifndef PREFERENCESDLG_H
#define PREFERENCESDLG_H

#include <QSettings>
#include <QSet>

#include "ui_preferences.h"

#include "common.h"
#include "uservideodlg.h"

#if QT_VERSION >= 0x050000
// in Qt 5 "General" section name is reserved :(
#define SETTINGS_GENERAL_NICKNAME                   "nickname"
#define SETTINGS_GENERAL_GENDER                     "gender"
#define SETTINGS_GENERAL_AUTOAWAY                   "auto-away"
#define SETTINGS_GENERAL_PUSHTOTALK                 "push-to-talk"
#define SETTINGS_GENERAL_PUSHTOTALK_KEY             "ptt-key"
#define SETTINGS_GENERAL_VOICEACTIVATED             "voice-activated"
#define SETTINGS_GENERAL_STATUSMESSAGE              "statusmsg"
#define SETTINGS_GENERAL_AUDIOSTORAGE_MODE          "audio-storage-mode"
#define SETTINGS_GENERAL_AUDIOSTORAGE_FOLDER        "audio-storage-folder"
#define SETTINGS_GENERAL_AUDIOSTORAGE_FILEFORMAT    "audio-storage-format"
#define SETTINGS_GENERAL_STREAMAUDIO_FILENAME       "stream-audio-file"
#define SETTINGS_GENERAL_STREAMAUDIO_LOOP           "stream-audio-file-loop"
#define SETTINGS_GENERAL_STREAMMEDIA_FILENAME       "stream-media-filename"
#define SETTINGS_GENERAL_STREAMMEDIA_AUDIO          "stream-media-audio"
#define SETTINGS_GENERAL_STREAMMEDIA_VIDEO          "stream-media-video"
#define SETTINGS_GENERAL_STREAMMEDIA_LOOP           "stream-media-file-loop"
#define SETTINGS_GENERAL_THEORA_QUALITY             "codec/theora-quality"
#define SETTINGS_GENERAL_THEORA_BITRATE             "codec/theora-bitrate"
#define SETTINGS_GENERAL_DESKTOPSHARE_MODE          "desktop-share-mode"
#define SETTINGS_GENERAL_DESKTOPSHARE_BMPMODE       "desktop-bmp-mode"
#define SETTINGS_GENERAL_DESKTOPSHARE_INTERVAL      "desktop-share-interval"
#define SETTINGS_GENERAL_DESKTOPSHARE_CURSOR        "desktop-cursor"
#define SETTINGS_GENERAL_DESKTOPSHARE_WINDOWTITLE   "desktop-window-title"
#define SETTINGS_GENERAL_LAST_DIRECTORY             "last-directory"
#else
#define SETTINGS_GENERAL_NICKNAME                   "general/nickname"
#define SETTINGS_GENERAL_GENDER                     "general/gender"
#define SETTINGS_GENERAL_AUTOAWAY                   "general/auto-away"
#define SETTINGS_GENERAL_PUSHTOTALK                 "general/push-to-talk"
#define SETTINGS_GENERAL_PUSHTOTALK_KEY             "general/ptt-key"
#define SETTINGS_GENERAL_VOICEACTIVATED             "general/voice-activated"
#define SETTINGS_GENERAL_STATUSMESSAGE              "general/statusmsg"
#define SETTINGS_GENERAL_AUDIOSTORAGE_MODE          "general/audio-storage-mode"
#define SETTINGS_GENERAL_AUDIOSTORAGE_FOLDER        "general/audio-storage-folder"
#define SETTINGS_GENERAL_AUDIOSTORAGE_FILEFORMAT    "general/audio-storage-format"
#define SETTINGS_GENERAL_STREAMAUDIO_FILENAME       "general/stream-audio-file"
#define SETTINGS_GENERAL_STREAMAUDIO_LOOP           "general/stream-audio-file-loop"
#define SETTINGS_GENERAL_STREAMMEDIA_FILENAME       "general/stream-media-filename"
#define SETTINGS_GENERAL_STREAMMEDIA_AUDIO          "general/stream-media-audio"
#define SETTINGS_GENERAL_STREAMMEDIA_VIDEO          "general/stream-media-video"
#define SETTINGS_GENERAL_STREAMMEDIA_LOOP           "general/stream-media-file-loop"
#define SETTINGS_GENERAL_THEORA_QUALITY             "general/codec/theora-quality"
#define SETTINGS_GENERAL_THEORA_BITRATE             "general/codec/theora-bitrate"
#define SETTINGS_GENERAL_DESKTOPSHARE_MODE          "general/desktop-share-mode"
#define SETTINGS_GENERAL_DESKTOPSHARE_BMPMODE       "general/desktop-bmp-mode"
#define SETTINGS_GENERAL_DESKTOPSHARE_INTERVAL      "general/desktop-share-interval"
#define SETTINGS_GENERAL_DESKTOPSHARE_CURSOR        "general/desktop-cursor"
#define SETTINGS_GENERAL_DESKTOPSHARE_WINDOWTITLE   "general/desktop-window-title"
#define SETTINGS_GENERAL_LAST_DIRECTORY             "general/last-directory"
#endif /* QT_VERSION */

#define SETTINGS_DISPLAY_MESSAGEPOPUP               "display/msgpopup"
#define SETTINGS_DISPLAY_VIDEOPOPUP                 "display/videopopup"
#define SETTINGS_DISPLAY_VIDEORETURNTOGRID          "display/videoreturntogrid"
#define SETTINGS_DISPLAY_VIDEORETURNTOGRID_DEFAULT  true
#define SETTINGS_DISPLAY_DESKTOPPOPUP               "display/desktoppopup"
#define SETTINGS_DISPLAY_VIDEOTEXT_WIDTH            "display/videotext-pane-width"
#define SETTINGS_DISPLAY_VIDEOTEXT_WIDTH_DEFAULT    30.0
#define SETTINGS_DISPLAY_VIDEOTEXT_HEIGHT           "display/videotext-pane-height"
#define SETTINGS_DISPLAY_VIDEOTEXT_HEIGHT_DEFAULT   10.0
#define SETTINGS_DISPLAY_VIDEOTEXT_FGCOLOR          "display/videotext-pane-fgcolor"
#define SETTINGS_DISPLAY_VIDEOTEXT_FGCOLOR_DEFAULT  QColor(Qt::white)
#define SETTINGS_DISPLAY_VIDEOTEXT_BGCOLOR          "display/videotext-pane-bgcolor"
#define SETTINGS_DISPLAY_VIDEOTEXT_BGCOLOR_DEFAULT  QColor(Qt::blue)
#define SETTINGS_DISPLAY_VIDEOTEXT                  "display/videotext-pane"
#define SETTINGS_DISPLAY_VIDEOTEXT_DEFAULT          (VIDTEXT_SHOW_NICKNAME | VIDTEXT_POSITION_BOTTOMRIGHT)
#define SETTINGS_DISPLAY_VIDEOTEXT_SHOW             "display/videotext-pane-show"
#define SETTINGS_DISPLAY_TRAYMINIMIZE               "display/trayminimize"
#define SETTINGS_DISPLAY_ALWAYSONTOP                "display/alwaysontop"
#define SETTINGS_DISPLAY_USERSCOUNT                 "display/userscount"
#define SETTINGS_DISPLAY_MSGTIMESTAMP               "display/msgtimestamp"
#define SETTINGS_DISPLAY_STARTMINIMIZED             "display/startminimized"
#define SETTINGS_DISPLAY_WINDOWPOS                  "display/windowposition"
#define SETTINGS_DISPLAY_SPLITTER                   "display/splitter"
#define SETTINGS_DISPLAY_DESKTOPSPLITTER            "display/desktop-splitter"
#define SETTINGS_DISPLAY_FILESHEADER                "display/filesheader"
#define SETTINGS_DISPLAY_LOGSTATUSBAR               "display/log-statusbar"
#define SETTINGS_DISPLAY_FREESERVERS                "display/showfreeservers"
#define SETTINGS_DISPLAY_LANGUAGE                   "display/language"
#define SETTINGS_DISPLAY_APPUPDATE                  "display/check-appupdate"
#define SETTINGS_DISPLAY_MAX_STRING                 "display/channel-max-string"
#define SETTINGS_DISPLAY_MAX_STRING_DEFAULT         TT_STRLEN

#define SETTINGS_CONNECTION_AUTOCONNECT             "connection/autoconnect"
#define SETTINGS_CONNECTION_RECONNECT               "connection/reconnect"
#define SETTINGS_CONNECTION_AUTOJOIN                "connection/autojoin"
#define SETTINGS_CONNECTION_QUERYMAXPAYLOAD         "connection/query-max-payload"
#define SETTINGS_CONNECTION_P2PAUDIO                "connection/p2paudio"
#define SETTINGS_CONNECTION_P2PVIDEO                "connection/p2pvideo"
#define SETTINGS_CONNECTION_SUBSCRIBE_USERMSG       "connection/subscribe-usermsg"
#define SETTINGS_CONNECTION_SUBSCRIBE_CHANNELMSG    "connection/subscribe-channelmsg"
#define SETTINGS_CONNECTION_SUBSCRIBE_BROADCASTMSG  "connection/subscribe-broadcastmsg"
#define SETTINGS_CONNECTION_SUBSCRIBE_AUDIO         "connection/subscribe-audio"
#define SETTINGS_CONNECTION_SUBSCRIBE_VIDEO         "connection/subscribe-video"
#define SETTINGS_CONNECTION_SUBSCRIBE_DESKTOP       "connection/subscribe-desktop"
#define SETTINGS_CONNECTION_TCPPORT                 "connection/localtcpport"
#define SETTINGS_CONNECTION_UDPPORT                 "connection/localudpport"

#define SETTINGS_SOUND_SOUNDSYSTEM                  "soundsystem/soundsystem"
#define SETTINGS_SOUND_INPUTDEVICE                  "soundsystem/inputdeviceid"
#define SETTINGS_SOUND_INPUTDEVICE_UID              "soundsystem/inputdeviceuid"
#define SETTINGS_SOUND_OUTPUTDEVICE                 "soundsystem/outputdeviceid"
#define SETTINGS_SOUND_OUTPUTDEVICE_UID             "soundsystem/outputdeviceuid"
#define SETTINGS_SOUND_MASTERVOLUME                 "soundsystem/mastervolume"
#define SETTINGS_SOUND_MICROPHONEGAIN               "soundsystem/microphonegain"
#define SETTINGS_SOUND_VOICEACTIVATIONLEVEL         "soundsystem/voice-activation-level"
#define SETTINGS_SOUND_3DPOSITIONING                "soundsystem/3dsound-positioning"
#define SETTINGS_SOUND_ECHOCANCEL                   "soundsystem/echocancellation"
#define SETTINGS_SOUND_DENOISING                    "soundsystem/denoising"
#define SETTINGS_SOUND_DUPLEXMODE                   "soundsystem/duplex-mode"

#define SETTINGS_SOUNDEVENT_NEWUSER                 "soundevents/newuser"
#define SETTINGS_SOUNDEVENT_REMOVEUSER              "soundevents/removeuser"
#define SETTINGS_SOUNDEVENT_SERVERLOST              "soundevents/server-lost"
#define SETTINGS_SOUNDEVENT_USERMSG                 "soundevents/user-msg"
#define SETTINGS_SOUNDEVENT_CHANNELMSG              "soundevents/channel-msg"
#define SETTINGS_SOUNDEVENT_HOTKEY                  "soundevents/hotkey-pressed"
#define SETTINGS_SOUNDEVENT_SILENCE                 "soundevents/channel-silent"
#define SETTINGS_SOUNDEVENT_NEWVIDEO                "soundevents/new-video-session"
#define SETTINGS_SOUNDEVENT_NEWDESKTOP              "soundevents/new-desktop-session"
#define SETTINGS_SOUNDEVENT_FILESUPD                "soundevents/files-updated"
#define SETTINGS_SOUNDEVENT_FILETXDONE              "soundevents/file-tx-done"
#define SETTINGS_SOUNDEVENT_QUESTIONMODE            "soundevents/question-mode"
#define SETTINGS_SOUNDEVENT_DESKTOPACCESS           "soundevents/desktop-access-request"

#define SETTINGS_SHORTCUTS_VOICEACTIVATION          "shortcuts/voiceactivation"
#define SETTINGS_SHORTCUTS_INCVOLUME                "shortcuts/volume-inc"
#define SETTINGS_SHORTCUTS_DECVOLUME                "shortcuts/volume-dec"
#define SETTINGS_SHORTCUTS_MUTEALL                  "shortcuts/muteall"
#define SETTINGS_SHORTCUTS_INCVOICEGAIN             "shortcuts/voicegain-inc"
#define SETTINGS_SHORTCUTS_DECVOICEGAIN             "shortcuts/voicegain-dec"
#define SETTINGS_SHORTCUTS_VIDEOTX                  "shortcuts/video-tx"

#define SETTINGS_VIDCAP_DEVICEID                    "videocapture/deviceid"
#define SETTINGS_VIDCAP_RESOLUTION                  "videocapture/resolution"
#define SETTINGS_VIDCAP_FPS                         "videocapture/fps"
#define SETTINGS_VIDCAP_FOURCC                      "videocapture/fourcc"
#define SETTINGS_VIDCAP_CODEC                       "videocapture/codec"
#define SETTINGS_VIDCAP_THEORA_QUALITY              "videocapture/codec/theora_quality"
#define SETTINGS_VIDCAP_THEORA_BITRATE              "videocapture/codec/theora_bitrate"

#define SETTINGS_LATESTHOST_HOSTADDR                "latesthosts/%1_hostaddr"
#define SETTINGS_LATESTHOST_TCPPORT                 "latesthosts/%1_tcpport"
#define SETTINGS_LATESTHOST_UDPPORT                 "latesthosts/%1_udpport"
#define SETTINGS_LATESTHOST_ENCRYPTED               "latesthosts/%1_encrypted"
#define SETTINGS_LATESTHOST_SRVPASSWD               "latesthosts/%1_srvpassword"
#define SETTINGS_LATESTHOST_USERNAME                "latesthosts/%1_username"
#define SETTINGS_LATESTHOST_PASSWORD                "latesthosts/%1_password"
#define SETTINGS_LATESTHOST_CHANNEL                 "latesthosts/%1_channel"
#define SETTINGS_LATESTHOST_CHANNELPASSWD           "latesthosts/%1_chanpassword"

#define SETTINGS_SERVERENTRIES_NAME                 "serverentries/%1_name"
#define SETTINGS_SERVERENTRIES_HOSTADDR             "serverentries/%1_hostaddr"
#define SETTINGS_SERVERENTRIES_TCPPORT              "serverentries/%1_tcpport"
#define SETTINGS_SERVERENTRIES_UDPPORT              "serverentries/%1_udpport"
#define SETTINGS_SERVERENTRIES_SRVPASSWD            "serverentries/%1_srvpassword"
#define SETTINGS_SERVERENTRIES_USERNAME             "serverentries/%1_username"
#define SETTINGS_SERVERENTRIES_PASSWORD             "serverentries/%1_password"
#define SETTINGS_SERVERENTRIES_CHANNEL              "serverentries/%1_channel"
#define SETTINGS_SERVERENTRIES_CHANNELPASSWD        "serverentries/%1_chanpassword"

#define SETTINGS_DESKTOPACCESS_HOSTADDR             "desktopaccess/%1_hostaddr"
#define SETTINGS_DESKTOPACCESS_TCPPORT              "desktopaccess/%1_tcpport"
#define SETTINGS_DESKTOPACCESS_CHANNEL              "desktopaccess/%1_%2_channel"
#define SETTINGS_DESKTOPACCESS_USERNAME             "desktopaccess/%1_%2_username"

#define SOUNDDEVICEID_DEFAULT  -2
#define SOUNDDEVICEID_NODEVICE -1

#define DEFAULT_VIDEO_WIDTH     320
#define DEFAULT_VIDEO_HEIGHT    240
#define DEFAULT_VIDEO_FPS       10
#define DEFAULT_VIDEO_FOURCC    FOURCC_RGB32
#define DEFAULT_THEORA_QUALITY  10
#define DEFAULT_THEORA_BITRATE  0

class PreferencesDlg : public QDialog
{
    Q_OBJECT

public:
    PreferencesDlg(QWidget * parent = 0);
    ~PreferencesDlg();

    enum
    {
        GENERAL_TAB,
        DISPLAY_TAB,
        CONNECTION_TAB,
        SOUND_TAB,
        SOUNDEVENTS_TAB,
        SHORTCUTS_TAB,
        VIDCAP_TAB
    };

private:
    Ui::PreferencesDlg ui;
    /* hotkey tab */
    hotkey_t m_hotkey;
    /* sound tab */
    void initDevices();
    SoundSystem getSoundSystem();
    void showDevices(SoundSystem snd);
    QVector<SoundDevice> m_inputdevices, m_outputdevices;
    /* sound events */
    bool getSoundFile(QString& filename);
    typedef QMap<HotKeyID, hotkey_t> hotkeys_t;
    hotkeys_t m_hotkeys;
    /* video tab */
    QVector<VideoCaptureDevice> m_videodevices;
    UserVideoDlg* m_uservideo;

    QSet<int> m_modtab;
    bool m_video_ready;

private slots:
    void slotTabChange(int index);
    void slotSaveChanges();
    void slotCancelChanges();
    void slotEnablePushToTalk(bool checked);
    void slotSetupHotkey();
    //display tab
    void slotLanguageChange(int index);
    void slotSelectVideoText();
    //connection tab
    void slotDesktopAccess();
    //sound tab
    void slotSoundSystemChange();
    void slotSoundInputChange(int index);
    void slotSoundOutputChange(int index);
    void slotSoundRestart();
    void slotSoundTestDevices(bool checked);
    void slotSoundDefaults();
    void slotUpdateSoundCheckBoxes();
    //sound events
    void slotEventNewUser();
    void slotEventRemoveUser();
    void slotEventServerLost();
    void slotEventUserTextMsg();
    void slotEventChannelTextMsg();
    void slotEventHotKey();
    void slotEventSilence();
    void slotEventNewVideo();
    void slotEventNewDesktop();
    void slotEventFilesUpdated();
    void slotEventFileTxDone();
    void slotEventQuestionMode();
    void slotEventDesktopAccess();
    //keyboard shortcuts
    void slotShortcutVoiceActivation(bool checked);
    void slotShortcutIncVolume(bool checked);
    void slotShortcutDecVolume(bool checked);
    void slotShortcutMuteAll(bool checked);
    void slotShortcutIncVoiceGain(bool checked);
    void slotShortcutDecVoiceGain(bool checked);
    void slotShortcutVideoTx(bool checked);
    //video tab
    void slotVideoCaptureDevChange(int index);
    void slotTestVideoFormat();
    void slotVideoQualityChanged(int value);
    void slotImageFormatChange(bool checked);
    void slotDefaultVideoSettings();

public slots:
    void slotNewVideoFrame(int userid, int frame_count);
};

#endif
