using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Globalization;
using System.Threading;
using BearWare;

namespace TeamTalk.NET
{

    public partial class MainForm : Form
    {
        TeamTalk4 ttclient;
        ChannelsView channels;
        UsersView users;
        FilesView files;

        Dictionary<int, MessageDlg> msgdialogs;
        Dictionary<int, VideoDlg> viddialogs;
        Dictionary<int, DesktopDlg> desktopdialogs;

        ClientStatistics statistics;

        Settings settings;

        IntPtr hShareWnd = IntPtr.Zero;
        BitmapFormat bmpShareFormat = BitmapFormat.BMP_NONE;

        //We want to pass the same bitmap reference every time so we don't
        //waste memory reallocating a new bitmap every time we call 
        //ttclient.GetUserVideoFrame
        Bitmap local_bitmap = null;

        public MainForm()
        {
            InitializeComponent();

            /* Set license information before creating the client instance */
            //TeamTalk4.SetLicenseInformation("", 0);

            msgdialogs = new Dictionary<int, MessageDlg>();
            viddialogs = new Dictionary<int, VideoDlg>();
            desktopdialogs = new Dictionary<int, DesktopDlg>();

            inputProgressBar.Minimum = SoundLevel.SOUND_VU_MIN;
            inputProgressBar.Maximum = SoundLevel.SOUND_VU_MAX;
            voiceactTrackBar.Minimum = SoundLevel.SOUND_VU_MIN;
            voiceactTrackBar.Maximum = SoundLevel.SOUND_VU_MAX;
            inputgainTrackBar.Minimum = SoundLevel.SOUND_GAIN_MIN;
            inputgainTrackBar.Maximum = SoundLevel.SOUND_GAIN_MAX;

            volumeTrackBar.Minimum = SoundLevel.SOUND_VOLUME_MIN;
            volumeTrackBar.Maximum = SoundLevel.SOUND_VOLUME_MAX;
            outputgainTrackBar.Minimum = SoundLevel.SOUND_GAIN_MIN;
            outputgainTrackBar.Maximum = SoundLevel.SOUND_GAIN_MAX;

            /* we pass 'false' to poll_events since we don't want to 
             * manually process events using ttclient.GetMessage */
            ttclient = new TeamTalk4(false);
            channels = new ChannelsView(ttclient, treeView1);
            users = new UsersView(ttclient, listView1);
            files = new FilesView(ttclient, filesListView);
            settings = new Settings();

            voiceactTrackBar.Value = ttclient.GetVoiceActivationLevel();
            inputgainTrackBar.Value = ttclient.GetSoundInputGainLevel();
            volumeTrackBar.Value = ttclient.GetSoundOutputVolume();
            outputgainTrackBar.Value = SoundLevel.SOUND_GAIN_DEFAULT;

            //get default devices
            ttclient.GetDefaultSoundDevices(out settings.sndinputid, out settings.sndoutputid);

            ttclient.OnConnectSuccess += new TeamTalk4.Connection(ttclient_OnConnectSuccess);
            ttclient.OnConnectFailed += new TeamTalk4.Connection(ttclient_OnConnectFailed);
            ttclient.OnConnectionLost += new TeamTalk4.Connection(ttclient_OnConnectionLost);
            ttclient.OnConnectionP2P += new TeamTalk4.ConnectionP2P(ttclient_OnConnectionP2P);

            ttclient.OnCmdProcessing += new TeamTalk4.CommandProcessing(ttclient_OnCmdProcessing);
            ttclient.OnCmdError += new TeamTalk4.CommandError(ttclient_OnCmdError);
            ttclient.OnCmdMyselfLoggedIn += new TeamTalk4.MyselfLoggedIn(ttclient_OnCmdMyselfLoggedIn);
            ttclient.OnCmdMyselfLoggedOut += new TeamTalk4.MyselfLoggedOut(ttclient_OnCmdMyselfLoggedOut);
            ttclient.OnCmdMyselfJoinedChannel += new TeamTalk4.MyselfJoinedChannel(ttclient_OnCmdMyselfJoinedChannel);
            ttclient.OnCmdMyselfLeftChannel += new TeamTalk4.MyselfLeftChannel(ttclient_OnCmdMyselfLeftChannel);
            ttclient.OnCmdUserLoggedIn += new TeamTalk4.UserAuth(ttclient_OnCmdUserLoggedIn);
            ttclient.OnCmdUserJoinedChannel += new TeamTalk4.UserUpdate(ttclient_OnCmdUserJoinedChannel);
            ttclient.OnCmdUserLeftChannel += new TeamTalk4.UserUpdate(ttclient_OnCmdUserLeftChannel);
            ttclient.OnCmdUserTextMessage += new TeamTalk4.UserTextMessage(ttclient_OnCmdUserTextMessage);
            ttclient.OnCmdChannelNew += new TeamTalk4.ChannelUpdate(ttclient_OnCmdChannelNew);
            ttclient.OnCmdChannelRemove += new TeamTalk4.ChannelUpdate(ttclient_OnCmdChannelRemove);

            ttclient.OnHotKeyToggle += new TeamTalk4.HotKeyToggle(ttclient_OnHotKeyToggle);
            ttclient.OnUserVideoFrame += new TeamTalk4.UserVideoFrame(ttclient_OnUserVideoFrame);
            ttclient.OnStreamAudioFileUser += new TeamTalk4.StreamUserAudioFile(ttclient_OnStreamAudioFileUser);
            ttclient.OnStreamAudioFileChannel += new TeamTalk4.StreamChannelAudioFile(ttclient_OnStreamAudioFileChannel);
            ttclient.OnStreamMediaFileChannel += new TeamTalk4.StreamChannelMediaFile(ttclient_OnStreamMediaFileChannel);
            ttclient.OnUserAudioFile += new TeamTalk4.UserAudioFile(ttclient_OnUserAudioFile);
            ttclient.OnUserAudioBlock += new TeamTalk4.NewAudioBlock(ttclient_OnUserAudioBlock);
            ttclient.OnUserDesktopInput += new TeamTalk4.UserDesktopInput(ttclient_OnUserDesktopInput);
            ttclient.OnFileTransfer += new TeamTalk4.FileTransferUpdate(ttclient_OnFileTransfer);
            ttclient.OnUserDesktopWindow += new TeamTalk4.NewDesktopWindow(ttclient_OnUserDesktopWindow);

            vumeterTimer.Enabled = true;
            timer1.Enabled = true;
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            this.CenterToScreen();

            UpdateControls();
        }

        void UpdateControls()
        {
            ClientFlag flags = ttclient.Flags;
            UserType myusertype = ttclient.UserType;
            ServerProperties srvprop;

            if ((flags & ClientFlag.CLIENT_CONNECTED) != ClientFlag.CLIENT_CONNECTED)
                ResetControls();

            ttclient.GetServerProperties(out srvprop);

            int userid = channels.GetSelectedUser();
            int channelid = channels.GetSelectedChannel();

            connectToolStripMenuItem.Text =
                ((flags & ClientFlag.CLIENT_CONNECTED) == ClientFlag.CLIENT_CONNECTED ||
                (flags & ClientFlag.CLIENT_CONNECTING) == ClientFlag.CLIENT_CONNECTING) ? "Disconnect" : "Connect";

            changeNicknameToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED;
            changeStatusModeToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED;
            enableDesktopSharingToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED;
            enableDesktopSharingToolStripMenuItem.Checked = (flags & ClientFlag.CLIENT_DESKTOP_ACTIVE) != ClientFlag.CLIENT_CLOSED;

            muteAllToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_SNDOUTPUT_READY) == ClientFlag.CLIENT_SNDOUTPUT_READY;
            muteAllToolStripMenuItem.Checked = (flags & ClientFlag.CLIENT_SNDOUTPUT_MUTE) == ClientFlag.CLIENT_SNDOUTPUT_MUTE;
            storeAudioToDiskToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED;
            storeAudioToDiskToolStripMenuItem.Checked = settings.audiofolder.Length > 0;

            User user;
            if (userid > 0 && ttclient.GetUser(userid, out user))
            {
                Debug.Assert((flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED);
                viewUserInformationToolStripMenuItem.Enabled = true;
                sendMessageToolStripMenuItem.Enabled = true;
                streamAudioFileToUserToolStripMenuItem.Enabled = 
                    (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN && 
                    ttclient.GetMyChannelID() != user.nChannelID;
                opDeOpToolStripMenuItem.Enabled = true;
                muteToolStripMenuItem.Enabled = true;
                muteToolStripMenuItem.Checked = (user.uUserState & UserState.USERSTATE_MUTE) == UserState.USERSTATE_MUTE;
                kickToolStripMenuItem.Enabled = true;
                kickAndBanToolStripMenuItem.Enabled = true;
                subscriptionsToolStripMenuItem.Enabled = true;

                userMessagesToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_USER_MSG) == Subscription.SUBSCRIBE_USER_MSG;
                channelMessagesToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_CHANNEL_MSG) == Subscription.SUBSCRIBE_CHANNEL_MSG;
                broadcastMessagesToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_BROADCAST_MSG) == Subscription.SUBSCRIBE_BROADCAST_MSG;
                audioToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_AUDIO) == Subscription.SUBSCRIBE_AUDIO;
                videoToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_VIDEO) == Subscription.SUBSCRIBE_VIDEO;
                desktopToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_DESKTOP) == Subscription.SUBSCRIBE_DESKTOP;
                interceptUserMessagesToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_USER_MSG) == Subscription.SUBSCRIBE_INTERCEPT_USER_MSG;
                interceptUserMessagesToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                interceptChannelMessagesToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_CHANNEL_MSG) == Subscription.SUBSCRIBE_INTERCEPT_CHANNEL_MSG;
                interceptChannelMessagesToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                interceptAudioToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_AUDIO) == Subscription.SUBSCRIBE_INTERCEPT_AUDIO;
                interceptAudioToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                interceptVideoToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_VIDEO) == Subscription.SUBSCRIBE_INTERCEPT_VIDEO;
                interceptVideoToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                interceptDesktopToolStripMenuItem.Checked = (user.uLocalSubscriptions & Subscription.SUBSCRIBE_INTERCEPT_DESKTOP) == Subscription.SUBSCRIBE_INTERCEPT_DESKTOP;
                interceptDesktopToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;

                advancedToolStripMenuItem.Enabled = true;

                Channel userchan;
                if (ttclient.GetChannel(user.nChannelID, out userchan))
                {
                    bool allowed = false;
                    foreach (int id in userchan.voiceUsers)
                    {
                        if (id == userid)
                            allowed = true;
                    }
                    allowVoiceTransmissionToolStripMenuItem.Checked = allowed;
                    allowed = false;
                    foreach (int id in userchan.videoUsers)
                    {
                        if (id == userid)
                            allowed = true;
                    }
                    allowVideoTransmissionToolStripMenuItem.Checked = allowed;
                    allowVoiceTransmissionToolStripMenuItem.Enabled = (userchan.uChannelType & ChannelType.CHANNEL_CLASSROOM) == ChannelType.CHANNEL_CLASSROOM;
                    allowVideoTransmissionToolStripMenuItem.Enabled = (userchan.uChannelType & ChannelType.CHANNEL_CLASSROOM) == ChannelType.CHANNEL_CLASSROOM;
                }
            }
            else
            {
                viewUserInformationToolStripMenuItem.Enabled = false;
                sendMessageToolStripMenuItem.Enabled = false;
                streamAudioFileToUserToolStripMenuItem.Enabled = false;
                opDeOpToolStripMenuItem.Enabled = false;
                muteToolStripMenuItem.Enabled = false;
                kickToolStripMenuItem.Enabled = false;
                kickAndBanToolStripMenuItem.Enabled = false;
                subscriptionsToolStripMenuItem.Enabled = false;
                advancedToolStripMenuItem.Enabled = false;
                allowVoiceTransmissionToolStripMenuItem.Enabled = false;
                allowVideoTransmissionToolStripMenuItem.Enabled = false;
            }

            Channel chan;
            if (channelid > 0 && ttclient.GetChannel(channelid, out chan))
            {
                Debug.Assert((flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED);
                joinChannelToolStripMenuItem.Enabled = true;
                joinNewChannelToolStripMenuItem.Enabled = (srvprop.uUserRights & UserRight.USERRIGHT_CHANNEL_CREATION) == UserRight.USERRIGHT_CHANNEL_CREATION;
                viewChannelInformationToolStripMenuItem.Enabled = true;
                createChannelToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                updateChannelToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN || 
                    ttclient.IsChannelOperator(ttclient.UserID, channelid);
                deleteChannelToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                streamAudioFileToChannelToolStripMenuItem.Enabled = ttclient.GetMyChannelID() == channelid;
                streamMediaFileToChannelToolStripMenuItem.Enabled = ttclient.GetMyChannelID() == channelid;
            }
            else
            {
                joinChannelToolStripMenuItem.Enabled = false;
                joinNewChannelToolStripMenuItem.Enabled =
                    (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED &&
                        (srvprop.uUserRights & UserRight.USERRIGHT_CHANNEL_CREATION) == UserRight.USERRIGHT_CHANNEL_CREATION;
                viewChannelInformationToolStripMenuItem.Enabled = false;
                createChannelToolStripMenuItem.Enabled =
                    (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED &&
                    (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
                updateChannelToolStripMenuItem.Enabled = false;
                deleteChannelToolStripMenuItem.Enabled = false;
                streamAudioFileToChannelToolStripMenuItem.Enabled = false;
                streamMediaFileToChannelToolStripMenuItem.Enabled = false;
            }

            listUserAccountsToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
            serverPropertiesToolStripMenuItem.Enabled = (flags & ClientFlag.CLIENT_AUTHORIZED) == ClientFlag.CLIENT_AUTHORIZED;
            saveConfigurationToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
            broadcastTestMessageToolStripMenuItem.Enabled = 
                (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN ||
                (srvprop.uUserRights & UserRight.USERRIGHT_CLIENT_BROADCAST) == UserRight.USERRIGHT_CLIENT_BROADCAST;
            serverStatisticsToolStripMenuItem.Enabled = (myusertype & UserType.USERTYPE_ADMIN) == UserType.USERTYPE_ADMIN;
        }

        void ResetControls()
        {
            listView1.Items.Clear();
            treeView1.Nodes.Clear();
        }

        void AddStatusMessage(string msg)
        {
            chatTextBox.AppendText("* " + msg + Environment.NewLine);
        }

        void ttclient_OnConnectSuccess()
        {
            AddStatusMessage("Connected to server");

            //the login dialog does its own comand error handling, so unregister event
            ttclient.OnCmdError -= ttclient_OnCmdError;

            LoginDlg dlg = new LoginDlg(ttclient, settings);
            dlg.ShowDialog();

            ttclient.OnCmdError += new TeamTalk4.CommandError(ttclient_OnCmdError);

            UpdateControls();
        }

        void ttclient_OnConnectFailed()
        {
            AddStatusMessage("Failed to connect");

            //close connnection
            ttclient.Disconnect();

            MessageBox.Show("Failed to connect");

            UpdateControls();
        }

        void ttclient_OnConnectionLost()
        {
            AddStatusMessage("Connection lost");

            //close connnection
            ttclient.Disconnect();

            MessageBox.Show("Connection dropped");

            UpdateControls();
        }

        void ttclient_OnConnectionP2P(int nUserID, bool bSuccess)
        {
            User user;
            if (ttclient.GetUser(nUserID, out user))
                if (bSuccess)
                    AddStatusMessage("Successful P2P connection to " + user.szNickname);
                else
                {
                    AddStatusMessage("Failed P2P connection to " + user.szNickname);

                    ServerProperties srvprop;
                    ttclient.GetServerProperties(out srvprop);
                    if ((srvprop.uUserRights & UserRight.USERRIGHT_FORWARD_AUDIO) == UserRight.USERRIGHT_NONE)
                        AddStatusMessage(user.szNickname + " will not be available for audio transmission");
                    if ((srvprop.uUserRights & UserRight.USERRIGHT_FORWARD_VIDEO) == UserRight.USERRIGHT_NONE)
                        AddStatusMessage(user.szNickname + " will not be available for video transmission");
                }
        }

        void ttclient_OnCmdProcessing(int nCmdID, bool bComplete)
        {
            if (bComplete)
            {
                UpdateControls();
            }
        }

        void ttclient_OnCmdError(ClientError nErrorNo, int nCmdID)
        {
            switch (nErrorNo)
            {
                case ClientError.CMDERR_INCORRECT_CHANNEL_PASSWORD :
                    MessageBox.Show("Invalid channel password", "Join Channel");
                    break;
                default:
                    MessageBox.Show(TeamTalk4.GetErrorMessage(nErrorNo), "Command Error");
                    break;
            }
        }

        void ttclient_OnCmdMyselfLoggedOut()
        {
            AddStatusMessage("Logged out of server");
        }

        void ttclient_OnCmdMyselfLoggedIn(int nMyUserID)
        {
            AddStatusMessage("Logged on to server successfully");

            UserAccount useraccount;
            if (ttclient.GetMyUserAccount(out useraccount))
            {
                string s = String.Format("User account information\r\n"+
                    "Username: {0}\r\n"+
                    "UserType: {1}\r\n"+
                    "UserData: {2}\r\n"+
                    "Note: {3}\r\n"+
                    "Initial Channel: {4}\r\n", 
                    useraccount.szUsername,
                    useraccount.uUserType.ToString(),
                    useraccount.nUserData, 
                    useraccount.szNote,
                    useraccount.szInitChannel);
                AddStatusMessage(s);
            }
        }

        void ttclient_OnCmdMyselfJoinedChannel(int nChannelID)
        {
            string path = "";
            if (ttclient.GetChannelPath(nChannelID, out path))
                AddStatusMessage("Joined channel " + path);

            if (settings.muxed_audio_file)
            {
                newMuxedAudioRecording();
            }
            UpdateControls();
        }

        void ttclient_OnCmdMyselfLeftChannel(int nChannelID)
        {
            string path = "";
            if (ttclient.GetChannelPath(nChannelID, out path))
                AddStatusMessage("Left channel " + path); 
            UpdateControls();
        }

        void ttclient_OnCmdUserLoggedIn(int nUserID)
        {
            //store audio to disk if an audio-folder has been specified
            if (!settings.muxed_audio_file && settings.audiofolder.Length > 0)
                ttclient.SetUserAudioFolder(nUserID, settings.audiofolder, "", settings.aff);
        }

        void ttclient_OnCmdUserJoinedChannel(int nUserID, int nChannelID)
        {
            User user;
            if (ttclient.GetUser(nUserID, out user))
            {
                //set default gain level for user (software gain volume)
                ttclient.SetUserGainLevel(nUserID, outputgainTrackBar.Value);
                if (user.nChannelID == ttclient.GetMyChannelID())
                    AddStatusMessage(user.szNickname + " joined channel");
            }
            UpdateControls();
        }

        void ttclient_OnCmdUserLeftChannel(int nUserID, int nChannelID)
        {
            User user;
            if (ttclient.GetUser(nUserID, out user)) //this will fail if 'UserRight.USERRIGHT_VIEW_ALL_USERS' is not set
            {
                if (nChannelID == ttclient.GetMyChannelID())
                    AddStatusMessage(user.szNickname + " left channel");
            }
            UpdateControls();
        }

        void ttclient_OnCmdUserTextMessage(int nUserID, int nMsgID)
        {
            TextMessage msg;
            if (!ttclient.GetTextMessage(nMsgID, true, out msg))
                return;

            switch (msg.nMsgType)
            {
                case TextMsgType.MSGTYPE_USER :
                    MessageDlg dlg;
                    if (msgdialogs.TryGetValue(msg.nFromUserID, out dlg))
                    {
                        dlg.NewMessage(msg);
                        if (!dlg.Visible)
                            dlg.Show();
                    }
                    else
                    {
                        dlg = new MessageDlg(ttclient, msg.nFromUserID);
                        dlg.FormClosed += new FormClosedEventHandler(dlg_MessageDlgClosed);
                        dlg.NewMessage(msg);
                        dlg.Show();
                        msgdialogs.Add(msg.nFromUserID, dlg);
                    }
                    break;
                case TextMsgType.MSGTYPE_CHANNEL :
                    User user;
                    if (ttclient.GetUser(msg.nFromUserID, out user))
                        chatTextBox.AppendText("<" + user.szNickname + "> " + msg.szMessage + Environment.NewLine);
                    break;
                case TextMsgType.MSGTYPE_BROADCAST :
                    if (ttclient.GetUser(msg.nFromUserID, out user))
                    {
                        MessageBox.Show("Broadcase message from " + user.szNickname + Environment.NewLine +
                                        msg.szMessage);
                    }
                    break;
            }
        }

        void ttclient_OnCmdChannelNew(int nChannelID)
        {
            UpdateControls();
        }

        void ttclient_OnCmdChannelRemove(int nChannelID)
        {
            UpdateControls();
        }

        void ttclient_OnHotKeyToggle(int nHotKeyID, bool bActive)
        {
            switch ((HotKey)nHotKeyID)
            {
                case HotKey.HOTKEY_PUSHTOTALK :
                    ttclient.EnableTransmission(TransmitType.TRANSMIT_AUDIO, bActive);
                    break;
            }
            Debug.WriteLine("HotKey " + nHotKeyID + " active " + bActive);
        }

        void ttclient_OnUserVideoFrame(int nUserID, int nFrameQueueSize)
        {
            if (nUserID == 0) //from local capture device
            {
#if TT_ZEROCOPY
                //Using zero-copy for displaying video images improves 
                //performance since the bitmap image-data is shared 
                //between the TeamTalk DLL file the .NET application.
                //It should, however, be used with causion since the
                //.NET application must ensure that the TeamTalk DLL
                //file hasn't released the memory which is referenced
                //in the .NET application.

                //Release bitmap resources since we're about to release
                //the memory shared between the .NET application and the
                //TeamTalk DLL.
                if (local_bitmap != null)
                    local_bitmap.Dispose();

                //Release shared memory
                ttclient.ReleaseUserVideoFrame(nUserID);

                //Acquire new image data. ENSURE to handle error situation
                //since we MUST NOT end up in a situation where the .NET
                //application is referencing shared memory.
                VideoFrame vidfrm = new VideoFrame();
                if (ttclient.AcquireUserVideoFrame(nUserID, ref vidfrm))
                {
                    PixelFormat pixelformat = PixelFormat.Format32bppRgb;
                    local_bitmap = new Bitmap(vidfrm.nWidth, vidfrm.nHeight,
                                           4 * vidfrm.nWidth, pixelformat,
                                           vidfrm.frameBuffer);
                    vidlocalPictureBox.Image = local_bitmap;
                }
                else
                {
                    //Failure situation. Set image to NULL so we don't 
                    //get an access violation exception by referencing
                    //released memory.
                    vidlocalPictureBox.Image = null;
                }
#else
                ttclient.GetUserVideoFrame(nUserID, ref local_bitmap);
                vidlocalPictureBox.Image = local_bitmap;
#endif
            }
            else
            {
                VideoDlg dlg;
                if (!viddialogs.TryGetValue(nUserID, out dlg))
                {
                    //Local video is 'nUserID' = 0;
                    dlg = new VideoDlg(ttclient, nUserID);
                    viddialogs.Add(nUserID, dlg);
                    dlg.FormClosed += new FormClosedEventHandler(videodlg_FormClosed);
                    dlg.Show();
                }
            }
        }

        void ttclient_OnStreamAudioFileUser(int nUserID, AudioFileStatus nStatus)
        {
            User user;
            if(!ttclient.GetUser(nUserID, out user))
                return;

            switch (nStatus)
            {
                case AudioFileStatus.AFS_ABORTED:
                    AddStatusMessage("Aborted audio stream to " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_ERROR:
                    AddStatusMessage("Error in audio stream to " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_FINISHED:
                    AddStatusMessage("Finished audio stream to " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_STARTED:
                    AddStatusMessage("Started audio stream to " + user.szNickname);
                    break;
            }
            streamAudioFileToUserToolStripMenuItem.Checked = nStatus == AudioFileStatus.AFS_STARTED;
        }

        void ttclient_OnStreamAudioFileChannel(int nChannelID, AudioFileStatus nStatus)
        {
            string chanpath = "";
            ttclient.GetChannelPath(nChannelID, out chanpath);

            switch (nStatus)
            {
                case AudioFileStatus.AFS_ABORTED:
                    AddStatusMessage("Aborted audio stream to " + chanpath);
                    break;
                case AudioFileStatus.AFS_ERROR:
                    AddStatusMessage("Error in audio stream to " + chanpath);
                    break;
                case AudioFileStatus.AFS_FINISHED:
                    AddStatusMessage("Finished audio stream to " + chanpath);
                    break;
                case AudioFileStatus.AFS_STARTED:
                    AddStatusMessage("Started audio stream to " + chanpath);
                    break;
            }
            streamAudioFileToChannelToolStripMenuItem.Checked = nStatus == AudioFileStatus.AFS_STARTED;
        }

        void ttclient_OnUserAudioFile(int nUserID, AudioFileStatus nStatus)
        {
            User user;
            ttclient.GetUser(nUserID, out user);
            switch (nStatus)
            {
                case AudioFileStatus.AFS_STARTED :
                    AddStatusMessage("Start audio file for " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_FINISHED :
                    AddStatusMessage("Finished audio file for " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_ERROR :
                    AddStatusMessage("Error writing audio file for " + user.szNickname);
                    break;
                case AudioFileStatus.AFS_ABORTED :
                    AddStatusMessage("Aborted audio file for " + user.szNickname);
                    break;
            }
        }

        void ttclient_OnUserAudioBlock(int nUserID)
        {
            AudioBlock block = new AudioBlock();
            if(ttclient.AcquireUserAudioBlock(nUserID, ref block))
            {
                ttclient.ReleaseUserAudioBlock(nUserID);
            }
        }

        void ttclient_OnUserDesktopWindow(int nUserID, int nSessionID)
        {
            DesktopDlg dlg;
            if (!desktopdialogs.TryGetValue(nUserID, out dlg))
            {
                //Local video is 'nUserID' = 0;
                dlg = new DesktopDlg(ttclient, nUserID);
                desktopdialogs.Add(nUserID, dlg);
                dlg.FormClosed += new FormClosedEventHandler(desktopdlg_FormClosed);
                dlg.Show();
            }
        }

        private void connectToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ClientFlag flags = ttclient.GetFlags();
            if ((flags & ClientFlag.CLIENT_CONNECTED) == ClientFlag.CLIENT_CONNECTED ||
                (flags & ClientFlag.CLIENT_CONNECTING) == ClientFlag.CLIENT_CONNECTING)
            {
                ttclient.Disconnect();
                UpdateControls();
                return;
            }

            ConnectDlg dlg = new ConnectDlg(ttclient, settings);
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                if ((flags & ClientFlag.CLIENT_SNDINPUT_READY) != ClientFlag.CLIENT_SNDINPUT_READY &&
                    !ttclient.InitSoundInputDevice(settings.sndinputid))
                    MessageBox.Show("Failed to initialize sound input device");
                if ((flags & ClientFlag.CLIENT_SNDOUTPUT_READY) != ClientFlag.CLIENT_SNDOUTPUT_READY &&
                    !ttclient.InitSoundOutputDevice(settings.sndoutputid))
                    MessageBox.Show("Failed to initialize sound output device");
            }
            UpdateControls();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            PreferencesDlg dlg = new PreferencesDlg(ttclient, settings);
            dlg.ShowDialog();
        }

        private void joinChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            if (channelid <= 0)
                return;
            //check if password protected
            Channel chan;
            if (!ttclient.GetChannel(channelid, out chan))
                return;
            string passwd = "";
            if (chan.bPassword)
                passwd = InputBox.Get("Join channel", "Password");
            ttclient.DoJoinChannelByID(channelid, passwd);
        }


        private void joinNewChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            if (channelid == 0)
                channelid = ttclient.GetRootChannelID(); //make the root parent if nothing is selected

            ServerProperties prop;
            ttclient.GetServerProperties(out prop);

            //check if server allows channels to be created dynamically
            if ((prop.uUserRights & UserRight.USERRIGHT_CHANNEL_CREATION) != UserRight.USERRIGHT_CHANNEL_CREATION)
            {
                MessageBox.Show("You are not administrator and server doesn't allow users to create channels", "Create Channel");
                return;
            }

            ChannelDlg dlg = new ChannelDlg(ttclient, ChannelDlgType.JOIN_CHANNEL, 0, channelid);
            dlg.ShowDialog();
        }

        private void vumeterTimer_Tick(object sender, EventArgs e)
        {
            inputProgressBar.Value = ttclient.GetSoundInputLevel();
        }

        private void voiceactTrackBar_ValueChanged(object sender, EventArgs e)
        {
            ttclient.SetVoiceActivationLevel(voiceactTrackBar.Value);
        }

        private void inputgainTrackBar_ValueChanged(object sender, EventArgs e)
        {
            ttclient.SetSoundInputGainLevel(inputgainTrackBar.Value);
        }

        private void volumeTrackBar_ValueChanged(object sender, EventArgs e)
        {
            ttclient.SetSoundOutputVolume(volumeTrackBar.Value);
        }

        private void outputgainTrackBar_ValueChanged(object sender, EventArgs e)
        {
            int channelid = ttclient.GetMyChannelID();
            if (channelid <= 0)
                return;

            //Here we just gain the user for all users. We could also use
            //SetUserVolume or SetUserGainLevel on individual users.
            int[] userids;
            if (ttclient.GetChannelUsers(channelid, out userids))
            {
                foreach (int userid in userids)
                    ttclient.SetUserGainLevel(userid, outputgainTrackBar.Value);
            }
        }

        private void treeView1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            UpdateControls();
        }

        private void chanmsgButton_Click(object sender, EventArgs e)
        {
            if (chanmsgTextBox.Text.Length > 0)
            {
                TextMessage msg;
                msg.nMsgType = TextMsgType.MSGTYPE_CHANNEL;
                msg.nChannelID = ttclient.GetMyChannelID();
                msg.nFromUserID = ttclient.GetMyUserID();
                msg.szFromUsername = ""; //not required
                msg.nToUserID = 0;
                msg.szMessage = chanmsgTextBox.Text;
                chanmsgTextBox.Text = "";
                ttclient.DoTextMessage(msg);
            }
        }

        private void enablePushToTalkToolStripMenuItem_Click(object sender, EventArgs e)
        {

        }

        private void viewChannelInformationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            ChannelDlg dlg = new ChannelDlg(ttclient, ChannelDlgType.VIEW_CHANNEL, channelid, 0);
            dlg.ShowDialog();
        }

        private void createChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            if (channelid == 0)
                channelid = ttclient.GetRootChannelID(); //make the root parent if nothing is selected

            ServerProperties prop;
            ttclient.GetServerProperties(out prop);

            //only administrator users can create a channel without joining them.
            UserType usertype = ttclient.GetMyUserType();
            if ((usertype & UserType.USERTYPE_ADMIN) != UserType.USERTYPE_ADMIN)
            {
                MessageBox.Show("Only administrators can create channels. Use 'Join new channel' instead", "Create Channel");
                return;
            }

            ChannelDlg dlg = new ChannelDlg(ttclient, ChannelDlgType.CREATE_CHANNEL, 0, channelid);
            dlg.ShowDialog();
        }

        private void updateChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            ChannelDlg dlg = new ChannelDlg(ttclient, ChannelDlgType.UPDATE_CHANNEL, channelid, 0);
            dlg.ShowDialog();
        }

        private void deleteChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int channelid = channels.GetSelectedChannel();
            string name = "";
            ttclient.GetChannelPath(channelid, out name);
            if (MessageBox.Show("Delete channel: " + name, "Delete Channel", MessageBoxButtons.YesNo) == DialogResult.Yes)
                ttclient.DoRemoveChannel(channelid);
        }

        private void sendMessageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int userid = channels.GetSelectedUser();
            MessageDlg dlg;
            if (msgdialogs.TryGetValue(userid, out dlg))
                dlg.Show();
            else
            {
                dlg = new MessageDlg(ttclient, userid);
                dlg.FormClosed += new FormClosedEventHandler(dlg_MessageDlgClosed);
                dlg.Show();
                msgdialogs.Add(userid, dlg);
            }
        }

        void dlg_MessageDlgClosed(object sender, FormClosedEventArgs e)
        {
            foreach (KeyValuePair<int, MessageDlg> dlg in msgdialogs)
            {
                if (dlg.Value == sender)
                {
                    msgdialogs.Remove(dlg.Key);
                    break;
                }
            }
        }

        void videodlg_FormClosed(object sender, FormClosedEventArgs e)
        {
            foreach (KeyValuePair<int, VideoDlg> dlg in viddialogs)
            {
                if (dlg.Value == sender)
                {
                    viddialogs.Remove(dlg.Key);
                    break;
                }
            }
        }

        void desktopdlg_FormClosed(object sender, FormClosedEventArgs e)
        {
            foreach (KeyValuePair<int, DesktopDlg> dlg in desktopdialogs)
            {
                if (dlg.Value == sender)
                {
                    desktopdialogs.Remove(dlg.Key);
                    break;
                }
            }
        }

        private void changeNicknameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string nickname = InputBox.Get("Change Nickname", "Nickname");
            ttclient.DoChangeNickname(nickname);
        }

        private void changeStatusModeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string msg = InputBox.Get("Change Status Mode", "Status Message");
            ttclient.DoChangeStatus(0, msg);
        }

        private void opDeOpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            if (!ttclient.GetUser(channels.GetSelectedUser(), out user))
                return;

            ttclient.DoChannelOp(user.nUserID, user.nChannelID, 
                !ttclient.IsChannelOperator(user.nUserID, user.nChannelID));
        }

        private void kickToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            if (!ttclient.GetUser(channels.GetSelectedUser(), out user))
                return;
            //pass 0 as 'nChannelID' to kick from server instead of channel
            ttclient.DoKickUser(user.nUserID, user.nChannelID);
        }

        private void kickAndBanToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            if (!ttclient.GetUser(channels.GetSelectedUser(), out user))
                return;
            //only admins can do this.
            ttclient.DoBanUser(user.nUserID);
            ttclient.DoKickUser(user.nUserID, user.nChannelID);
        }

        private void listUserAccountsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //only admins can do this.
            UserAccountsDlg dlg = new UserAccountsDlg(ttclient);
            dlg.ShowDialog();
        }

        private void serverPropertiesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ServerPropertiesDlg dlg = new ServerPropertiesDlg(ttclient);
            dlg.ShowDialog();
        }

        private void saveConfigurationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ttclient.DoSaveConfig();
        }

        private void muteAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ttclient.SetSoundOutputMute(muteAllToolStripMenuItem.Checked);
        }

        private void subscribeCommon(int userid, Subscription sub)
        {
            User user;
            if (!ttclient.GetUser(userid, out user))
                return;

            if ((user.uLocalSubscriptions & sub) == sub)
                ttclient.DoUnsubscribe(userid, sub);
            else
                ttclient.DoSubscribe(userid, sub);
        }

        private void userMessagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_USER_MSG);
        }

        private void channelMessagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_CHANNEL_MSG);
        }

        private void broadcastMessagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_BROADCAST_MSG);
        }

        private void audioToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_AUDIO);
        }

        private void videoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_VIDEO);
        }

        private void desktopToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_DESKTOP);
        }

        private void desktopAccessToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_DESKTOPINPUT);
        }

        private void allowDesktopAccessToolStripMenuItem_Click(object sender, EventArgs e)
        {
            desktopAccessToolStripMenuItem_Click(sender, e);
        }

        private void interceptUserMessagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_INTERCEPT_USER_MSG);
        }

        private void interceptChannelMessagesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_INTERCEPT_CHANNEL_MSG);
        }

        private void interceptAudioToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_INTERCEPT_AUDIO);
        }

        private void interceptVideoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_INTERCEPT_VIDEO);
        }

        private void interceptDesktopToolStripMenuItem_Click(object sender, EventArgs e)
        {
            subscribeCommon(channels.GetSelectedUser(), Subscription.SUBSCRIBE_INTERCEPT_DESKTOP);
        }

        private void chanmsgTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
                chanmsgButton_Click(sender, e);
        }

        private void viewUserInformationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UserInfoDlg dlg = new UserInfoDlg(ttclient, channels.GetSelectedUser());
            dlg.ShowDialog();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            ClientStatistics tmp;
            if (!ttclient.GetStatistics(out tmp))
                return;

            double totalrx = (tmp.nUdpBytesRecv - statistics.nUdpBytesRecv) / 1024;
            double totaltx = (tmp.nUdpBytesSent - statistics.nUdpBytesSent) / 1024;
            double voicerx = (tmp.nVoiceBytesRecv - statistics.nVoiceBytesRecv) / 1024;
            double voicetx = (tmp.nVoiceBytesSent - statistics.nVoiceBytesSent) / 1024;
            double videorx = (tmp.nVideoBytesRecv - statistics.nVideoBytesRecv) / 1024;
            double videotx = (tmp.nVideoBytesSent - statistics.nVideoBytesSent) / 1024;
            double desktoprx = (tmp.nDesktopBytesRecv - statistics.nDesktopBytesRecv) / 1024;
            double desktoptx = (tmp.nDesktopBytesSent - statistics.nDesktopBytesSent) / 1024;
            statistics = tmp;

            //toolStripStatusLabel1.Text = String.Format("RX: {0:F} TX: {1:F} VoiceRX: {2:F} VoiceTX {3:F} VideoRX {4:F} VideoTX {5:F}",
            //                totalrx, totaltx, voicerx, voicetx, videorx, videotx);

            toolStripStatusLabel1.Text = String.Format("RX/TX Total: {0:F}/{1:F} Audio: {2:F}/{3:F} Video: {4:F}/{5:F} Desktop: {4:F}/{5:F} in KBytes",
                            totalrx, totaltx, voicerx, voicetx, videorx, videotx, desktoprx, desktoptx);
        }

        private void storeAudioToDiskToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (storeAudioToDiskToolStripMenuItem.Checked)
            {
                AudioStorageDlg dlg = new AudioStorageDlg(settings);
                if (dlg.ShowDialog() != DialogResult.OK)
                    return;

                if (settings.muxed_audio_file)
                {
                    if (ttclient.ChannelID > 0)
                        newMuxedAudioRecording();
                }
                else
                {
                    List<int> userids = users.GetUsers();
                    foreach (int id in userids)
                        ttclient.SetUserAudioFolder(id, settings.audiofolder, "", settings.aff);
                    //To store audio in other channels, log in as admin, call DoSubscribe(.,SUBSCRIBE_INTERCEPT_AUDIO) 
                    //and set SetUserAudioFolder() on the user
                }
            }
            else
            {
                settings.audiofolder = "";
                settings.muxed_audio_file = false;
                settings.aff = AudioFileFormat.AFF_NONE;
                //clear if single file
                ttclient.StopRecordingMuxedAudioFile();
                //clear if separate files
                List<int> userids = users.GetUsers();
                foreach (int id in userids)
                    ttclient.SetUserAudioFolder(id, "", "", settings.aff);
            }

            UpdateControls();
        }

        private void newMuxedAudioRecording()
        {
            ttclient.StopRecordingMuxedAudioFile();

            Channel chan;
            if (!ttclient.GetChannel(ttclient.ChannelID, out chan))
            {
                MessageBox.Show("Must be in a channel to start muxed audio recording");
                return;
            }

            //generate a filename for the new mux file.
            CultureInfo currentCulture = Thread.CurrentThread.CurrentCulture;
            Thread.CurrentThread.CurrentCulture = CultureInfo.CreateSpecificCulture("ja-JP");
            string timestamp = DateTime.Now.ToString();
            timestamp = timestamp.Replace(":", "");
            timestamp = timestamp.Replace("/", "");
            Thread.CurrentThread.CurrentCulture = currentCulture;
            string extension = "";
            switch (settings.aff)
            {
                case AudioFileFormat.AFF_MP3_64KBIT_FORMAT :
                case AudioFileFormat.AFF_MP3_128KBIT_FORMAT :
                case AudioFileFormat.AFF_MP3_256KBIT_FORMAT :
                    extension = ".mp3"; break;
                case AudioFileFormat.AFF_WAVE_FORMAT :
                    extension = ".wav";
                    break;
            }
            string filename = settings.audiofolder + "\\" + timestamp + " Conference" + extension;
            if (!ttclient.StartRecordingMuxedAudioFile(chan.codec, filename, settings.aff))
                MessageBox.Show("Failed to create muxed audio file: " + filename);
            else
                AddStatusMessage("Recording to " + filename);
        }

        List<ToolStripItem> contextmenuitems = new List<ToolStripItem>();
        ToolStripMenuItem menu;

        private void contextMenuStrip1_Opening(object sender, CancelEventArgs e)
        {
            int userid = channels.GetSelectedUser();
            if (userid > 0)
            {
                menu = usersToolStripMenuItem;
                foreach(ToolStripItem item in usersToolStripMenuItem.DropDownItems)
                    contextmenuitems.Add(item);
            }
            else
            {
                menu = channelsToolStripMenuItem;
                foreach(ToolStripItem item in channelsToolStripMenuItem.DropDownItems)
                    contextmenuitems.Add(item);
            }
            foreach (ToolStripItem item in contextmenuitems)
                channelsContextMenuStrip.Items.Add(item);
            e.Cancel = false;
        }

        private void contextMenuStrip1_Closed(object sender, ToolStripDropDownClosedEventArgs e)
        {
            foreach (ToolStripItem item in contextmenuitems)
                menu.DropDownItems.Add(item);
            contextmenuitems.Clear();
        }

        private void streamAudioFileToUserToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //An admin can stream outside his own channel. Default users cannot.

            //Note that it is not possible to stream an audio file to a user in the same
            //channel as 'myself'. A user can only have one active audio stream in a channel.
            int userid = channels.GetSelectedUser();
            if (userid <= 0)
                return;

            OpenFileDialog openFileDialog1 = new OpenFileDialog();

            openFileDialog1.RestoreDirectory = true;
            openFileDialog1.Filter = "Wav files (*.wav)|*.wav";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if (!ttclient.StartStreamingAudioFileToUser(userid, openFileDialog1.FileName))
                    MessageBox.Show("Failed to stream wave-file");
                //it can be stopped using:
                //ttclient.StopStreamingAudioFile(userid);
            }
        }

        private void streamAudioFileToChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            //A wavefile cannot be stream to a channel other than the one we're currently in.
            //A wavefile can, however, be streamed to a single user in another channel
            int chanid = ttclient.GetMyChannelID();
            if (chanid <= 0)
                return;

            if (streamAudioFileToChannelToolStripMenuItem.Checked)
            {
                ttclient.StopStreamingAudioFileToChannel(chanid);
                streamAudioFileToChannelToolStripMenuItem.Checked = false;
                return;
            }

            OpenFileDialog openFileDialog1 = new OpenFileDialog();

            openFileDialog1.RestoreDirectory = true;
            openFileDialog1.Filter = "Wav files (*.wav)|*.wav";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if (!ttclient.StartStreamingAudioFileToChannel(chanid, 
                    openFileDialog1.FileName))
                    MessageBox.Show("Failed to stream wave-file");
                else
                    streamAudioFileToChannelToolStripMenuItem.Checked = true;

                //it can be stopped using:
                //ttclient.StopStreamingAudioFile(0);
            }
        }

        private void broadcastTestMessageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TextMessage msg;
            msg.nMsgType = TextMsgType.MSGTYPE_BROADCAST;
            msg.nChannelID = 0;
            msg.nFromUserID = ttclient.GetMyUserID();
            msg.szFromUsername = ""; //not required
            msg.nToUserID = 0;
            msg.szMessage = InputBox.Get("Broadcast message", "Message");
            ttclient.DoTextMessage(msg);
        }

        private void filesContextMenuStrip_Opened(object sender, EventArgs e)
        {
            downloadFileToolStripMenuItem.Enabled = files.GetSelectedFile() > 0;
            uploadFileToolStripMenuItem.Enabled = files.GetSelectedChannel() > 0;
            deleteFileToolStripMenuItem.Enabled = files.GetSelectedFile() > 0;
        }

        private void uploadFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog1 = new OpenFileDialog();

            openFileDialog1.RestoreDirectory = true;
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ttclient.DoSendFile(files.GetSelectedChannel(), openFileDialog1.FileName);
            }
        }

        private void downloadFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int fileid = files.GetSelectedFile();
            int channelid = files.GetSelectedChannel();

            FileInfo file;
            if (!ttclient.GetChannelFileInfo(channelid, fileid, out file))
                return;

            SaveFileDialog saveFileDialog1 = new SaveFileDialog();
            saveFileDialog1.FileName = file.szFileName;
            saveFileDialog1.RestoreDirectory = true;

            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                ttclient.DoRecvFile(channelid, fileid, saveFileDialog1.FileName);
        }

        private void deleteFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ttclient.DoDeleteFile(files.GetSelectedChannel(), files.GetSelectedFile());
        }

        void ttclient_OnFileTransfer(int nTransferID, FileTransferStatus nStatus)
        {
            if (nStatus == FileTransferStatus.FILETRANSFER_STARTED)
                new FileTransferDlg(ttclient, nTransferID).Show();
        }

        private void vidtxCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if ((ttclient.GetFlags() & ClientFlag.CLIENT_VIDEO_READY) == ClientFlag.CLIENT_VIDEO_READY)
                ttclient.EnableTransmission(TransmitType.TRANSMIT_VIDEO,
                    vidtxCheckBox.Checked);
            else
            {
                MessageBox.Show("Video device hasn't been configured in Preferences");
                vidtxCheckBox.Checked = false;
            }
            UpdateControls();
        }

        private void pttCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            //Note that it's not advised to enable hotkeys when debugging.
            //The debugger gets REALLY slow
            if (pttCheckBox.Checked)
            {
                HotKeyDlg dlg = new HotKeyDlg(ttclient);
                dlg.ShowDialog();
                ttclient.HotKey_Register((int)HotKey.HOTKEY_PUSHTOTALK, dlg.keys.ToArray());
                //OnHotKeyToggle is called when hotkey become active
            }
            else
            {
                ttclient.HotKey_Unregister((int)HotKey.HOTKEY_PUSHTOTALK);
            }
            UpdateControls();
        }

        private void voxtxCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            voiceactTrackBar.Enabled = voxtxCheckBox.Checked;
            ttclient.EnableVoiceActivation(voxtxCheckBox.Checked);
        }

        private void muteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            if(ttclient.GetUser(channels.GetSelectedUser(), out user))
                ttclient.SetUserMute(user.nUserID,
                            (user.uUserState & UserState.USERSTATE_MUTE) != UserState.USERSTATE_MUTE);
        }

        private void allowVoiceTransmissionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            Channel chan;
            if (ttclient.GetUser(channels.GetSelectedUser(), out user) && 
                ttclient.GetChannel(user.nChannelID, out chan))
            {
                //store users who are currently allowed to transmit
                List<int> voiceUsers = new List<int>();
                foreach (int id in chan.voiceUsers)
                {
                    if (id > 0)
                        voiceUsers.Add(id);
                }
                //add or remove user
                if(allowVoiceTransmissionToolStripMenuItem.Checked)
                    voiceUsers.Add(user.nUserID);
                else
                    voiceUsers.Remove(user.nUserID);

                //issue channel-update command to update list of allowed transmissions
                if (voiceUsers.Count > TeamTalk4.TT_VOICEUSERS_MAX)
                    MessageBox.Show("Maximum users is " + TeamTalk4.TT_VOICEUSERS_MAX.ToString());
                else
                {
                    for (int i = 0; i < TeamTalk4.TT_VOICEUSERS_MAX; i++)
                    {
                        if (i < voiceUsers.Count)
                            chan.voiceUsers[i] = voiceUsers[i];
                        else
                            chan.voiceUsers[i] = 0;
                    }
                    ttclient.DoUpdateChannel(chan);
                }
            }
        }

        private void allowVideoTransmissionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            User user;
            Channel chan;
            if (ttclient.GetUser(channels.GetSelectedUser(), out user) &&
                ttclient.GetChannel(user.nChannelID, out chan))
            {
                //store users who are currently allowed to transmit
                List<int> videoUsers = new List<int>();
                foreach (int id in chan.videoUsers)
                {
                    if (id > 0)
                        videoUsers.Add(id);
                }
                //add or remove user
                if (allowVideoTransmissionToolStripMenuItem.Checked)
                    videoUsers.Add(user.nUserID);
                else
                    videoUsers.Remove(user.nUserID);

                //issue channel-update command to update list of allowed transmissions
                if (videoUsers.Count > TeamTalk4.TT_VIDEOUSERS_MAX)
                    MessageBox.Show("Maximum users is " + TeamTalk4.TT_VIDEOUSERS_MAX.ToString());
                else
                {
                    for (int i = 0; i < TeamTalk4.TT_VIDEOUSERS_MAX; i++)
                    {
                        if (i < videoUsers.Count)
                            chan.videoUsers[i] = videoUsers[i];
                        else
                            chan.videoUsers[i] = 0;
                    }
                    ttclient.DoUpdateChannel(chan);
                }
            }
        }

        private void serverStatisticsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new ServerStatsDlg(ttclient).ShowDialog();
        }

        private void treeView1_ItemDrag(object sender, ItemDragEventArgs e)
        {
            TreeNode item = (TreeNode)e.Item;
            if (item.ImageIndex == (int)ChannelsView.ImageIndex.USER)
                treeView1.DoDragDrop(item, DragDropEffects.Move);
        }

        private void treeView1_DragEnter(object sender, DragEventArgs e)
        {
            e.Effect = DragDropEffects.Move;
        }

        private void treeView1_DragDrop(object sender, DragEventArgs e)
        {
            Point p = ((TreeView)sender).PointToClient(new Point(e.X, e.Y));
            TreeNode targetNode = ((TreeView)sender).GetNodeAt(p);
            if (targetNode == null)
                return;
            int chanid = 0;
            switch (targetNode.ImageIndex)
            {
                case (int)ChannelsView.ImageIndex.CHANNEL:
                    chanid = (int)targetNode.Tag;
                    break;
                case (int)ChannelsView.ImageIndex.USER:
                    User user;
                    if(ttclient.GetUser((int)targetNode.Tag, out user))
                        chanid = user.nChannelID;
                    break;
            }
            if (chanid == 0)
                return;

            TreeNode sourceNode = (TreeNode)e.Data.GetData("System.Windows.Forms.TreeNode");
            ttclient.DoMoveUser((int)sourceNode.Tag, chanid);
        }

        private void enableDesktopSharingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if ((ttclient.Flags & ClientFlag.CLIENT_DESKTOP_ACTIVE) == ClientFlag.CLIENT_DESKTOP_ACTIVE)
            {
                ttclient.CloseDesktopWindow();
                UpdateControls();

                senddesktopTimer.Enabled = false;
                sendcursorTimer.Enabled = false;
                return;
            }

            DesktopShareDlg dlg = new DesktopShareDlg();
            if (dlg.ShowDialog() != DialogResult.OK)
                return;

            //'hShareWnd' == IntPtr.Zero means share active window
            hShareWnd = dlg.hShareWnd;
            bmpShareFormat = dlg.bmpformat;

            if (sendDesktopWindow())
            {
                if (dlg.update_interval > 0)
                {
                    senddesktopTimer.Interval = dlg.update_interval;
                    senddesktopTimer.Enabled = true;
                }
                //send desktop cursor every 50 msec
                if (dlg.share_cursor)
                {
                    sendcursorTimer.Interval = 50;
                    sendcursorTimer.Enabled = true;
                }
            }
            else
                MessageBox.Show("Failed to send shared window", "Desktop Sharing Error");

            UpdateControls();
        }

        bool sendDesktopWindow()
        {
            //figure out which window to share
            IntPtr hWnd;
            if (hShareWnd == IntPtr.Zero)
                hWnd = WindowsHelper.GetDesktopActiveHWND();
            else
                hWnd = hShareWnd;

            //don't try to send new bitmap if one is already being transmitted
            if ((ttclient.Flags & ClientFlag.CLIENT_TX_DESKTOP) == ClientFlag.CLIENT_CLOSED)
                return ttclient.SendDesktopWindowFromHWND(hWnd, bmpShareFormat,
                                                          DesktopProtocol.DESKTOPPROTOCOL_ZLIB_1)>0;
            return false;
        }

        private void senddesktopTimer_Tick(object sender, EventArgs e)
        {
            sendDesktopWindow();
        }

        private void sendcursorTimer_Tick(object sender, EventArgs e)
        {
            ttclient.SendDesktopCursorPosition(0, Cursor.Position.X, Cursor.Position.Y);
        }

        private void streamMediaFileToChannelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if ((ttclient.Flags & (ClientFlag.CLIENT_STREAM_AUDIO | ClientFlag.CLIENT_STREAM_VIDEO)) != ClientFlag.CLIENT_CLOSED)
            {
                ttclient.StopStreamingMediaFileToChannel();
                streamMediaFileToChannelToolStripMenuItem.Checked = false;
            }
            else
                new MediaFileDlg(ttclient).ShowDialog();
        }

        void ttclient_OnStreamMediaFileChannel(AudioFileStatus nStatus)
        {
            switch (nStatus)
            {
                case AudioFileStatus.AFS_ABORTED:
                    AddStatusMessage("Aborted audio stream to channel");
                    break;
                case AudioFileStatus.AFS_ERROR:
                    AddStatusMessage("Error in audio stream to channel");
                    break;
                case AudioFileStatus.AFS_FINISHED:
                    AddStatusMessage("Finished audio stream to channel");
                    break;
                case AudioFileStatus.AFS_STARTED:
                    AddStatusMessage("Started audio stream to channel");
                    break;
            }
            streamMediaFileToChannelToolStripMenuItem.Checked = nStatus == AudioFileStatus.AFS_STARTED;
        }

        void ttclient_OnUserDesktopInput(int nUserID)
        {
            DesktopInput[] inputs = null, outputs = null;
            if(!ttclient.GetUserDesktopInput(nUserID, out inputs))
                return;
            //assumes desktop input is received in TTKEYCODE format
            WindowsHelper.KeyTranslate(TTKeyTranslate.TTKEY_TTKEYCODE_TO_WINKEYCODE,
                                       inputs, out outputs);
            WindowsHelper.Execute(outputs);
        }
    }
}